/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package chess;

/**
 * Implements the relative history heuristic.
 * @author petero
 */
public class History {
    private int countSuccess[][];
    private int countFail[][];
    int numUpdates;

    public History() {
        countSuccess = new int[Piece.values().length][64];
        countFail = new int[Piece.values().length][64];
        for (Piece p : Piece.values()) {
            for (int sq = 0; sq < 64; sq++) {
                countSuccess[p.ordinal()][sq] = 0;
                countFail[p.ordinal()][sq] = 0;
            }
        }
        numUpdates = 0;
    }

    /** Record move as a success. */
    public final synchronized void addSuccess(Position pos, Move m) {
        Piece p = pos.getPiece(m.from);
        countSuccess[p.ordinal()][m.to]++;
        renormalizeIfNeeded();
    }

    /** Record move as a failure. */
    public final synchronized void addFail(Position pos, Move m) {
        Piece p = pos.getPiece(m.from);
        countFail[p.ordinal()][m.to]++;
        renormalizeIfNeeded();
    }

    /** Get a score between 0 and 49, depending of the success/fail ratio of the move. */
    public final synchronized int getHistScore(Position pos, Move m) {
        Piece p = pos.getPiece(m.from);
        int succ = countSuccess[p.ordinal()][m.to];
        int fail = countFail[p.ordinal()][m.to];
        if (succ + fail > 0) {
            return succ * 49 / (succ + fail);
        } else {
            return 0;
        }
    }

    /** Renormalize history data if needed, to avoid integer overflow. */
    private final void renormalizeIfNeeded() {
        numUpdates++;
        if (numUpdates >= 1000000000) {
            for (Piece p : Piece.values()) {
                for (int sq = 0; sq < 64; sq++) {
                    countSuccess[p.ordinal()][sq] /= 2;
                    countFail[p.ordinal()][sq] /= 2;
                }
            }
            numUpdates /= 2;
        }
    }
}
